/* skystreets.cpp
 * Main game stuff
 * Licensed under the Open Software License version 2.0
 * Coded by ReKleSS [rekless@fastmail.fm]
 */


#include "ss.h"
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include "SDL.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <memory.h>
#include <stdlib.h>
#include "levels.h"
#include "explosion.h"
#include "SDL_image.h"
#include "glm.h"
#include "audio.h"
#include "text.h"

#define TEXTURES 15

#define FORWARD 0
#define LEFT 1
#define RIGHT 2
#define UP 3
// {{{ Forward declarations
void InitGL();
void gDrawScene();
void MouseMoved(SDL_MouseMotionEvent mouse);
void LoadTextures();
bool WriteNum(SDL_Surface * target, int number);

void initMenu();
void mDrawScene();

void gRunEditor();
// }}}

// {{{ Global vars
int state;
int tick;
bool g_RMouse;
bool g_MMouse;
bool g_LMouse;
lvlpoint *e_edSelected = NULL;
float g_mousex, g_mousey, g_camPos;
float g_throttle;
float g_fuel;
float g_oxygen;
FILE *debug;

long lastCycle;
long currentTime;
long trackTime;
bool finished;

Uint32 g_time;
int frames;
GLuint textures[TEXTURES];
GLuint level;
bool keys[256];
struct actor player;
struct vector Target;
struct actor cursor;
Level current;
int fps = 0;
lvlpoint curPoint;
int e_curMaterial = 1;
audioSample explosion;
audioSample engine;
Text text;

bool stVert = 0;		//are we in the vertical stage?
int editorState;

int fontLists[11];

int shipList;

char basePath[] = { DATAPATH };
// }}}

void debugprint(char *message)
{
	fprintf(debug, message);
}

Uint32 AddFrames(Uint32 interval, void *param)
{
	fps = frames;
	frames = 0;
	tick++;
	g_throttle = 0;

	return interval;
}

SDL_Surface *screen;

void Reset()
{
	memset(&player, 0, sizeof(actor));
	player.z = 0.5f;
	player.y = 8.0f;
	current.curBlock = 0;
	g_fuel = current.maxFuel;
	g_oxygen = current.maxOxygen;
	trackTime = 0;
	finished = 0;
	engine.stop();
	explosion.stop();
}

int main(int argc, char *argv[])
{
	SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER);
	memset(&player, 0, sizeof(actor));
	SDLKey sym;
	srand(SDL_GetTicks());
	g_RMouse = false;
	g_MMouse = false;
	g_LMouse = false;
	g_camPos = 0.0f;
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	bool fullscreen = true;
	int nlvl = 0;
	// {{{ ugly command line handling stuff

	for (int i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {	//single character flags...
			case 'w':	//windowed mode
				fullscreen = false;
				break;
#ifndef WIN32
			case 'a':	//antialiasing!!!
				if (argv[i][2])
					setenv("__GL_FSAA_MODE",
					       &argv[i][2], 0);
				else
					setenv("__GL_FSAA_MODE", "4", 0);
				break;

#endif
			case 's':	//silent
				audioSample::disabled = 1;
				break;
			}
		} else
			nlvl = i;
	}
	// }}}
	screen =
	    SDL_SetVideoMode(800, 600, 16,
			     SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_OPENGL |
			     (fullscreen ? SDL_FULLSCREEN : 0));
	SDL_WM_SetCaption("SkyStreets v0.2", NULL);
	printf("Starting SkyStreets v0.2\n");

	player.x = 0;
	player.y = 8.0f;
	player.z = 0.5f;
	player.spdy = 0;
	state = MENU;
	player.exploding = false;

	Target.x = 0;
	Target.y = 0.1f;
	Target.z = -12.0f;

	fps = 0;
	tick = 0;
	memset(&curPoint, 0, sizeof(lvlpoint));
	curPoint.blx = -1;

	stVert = false;

	initParticles();

	lastCycle = 0;
	currentTime = SDL_GetTicks();

	char fullpath[64];
#ifndef WIN32
	FILE *file = fopen("gfx/ship.obj", "r");
	if (file == NULL) {	//try the other path
		printf("trying system path\n");
		strcpy(fullpath, basePath);
		strcat(fullpath, "gfx/ship.obj");
		file = fopen(fullpath, "r");
		if (file == NULL) {
			printf("Can't find data files; dying\n");
			return 1;
		}
	} else
#endif
		strcpy(basePath, "./");	//current dir

	InitGL();
	initMenu();
	text.init();
	printf("GL Init'd\n");

	strcpy(fullpath, basePath);
	strcat(fullpath, "audio/explosion.wav");
	if (explosion.load(fullpath) == -1)
		printf("explosion.wav didn't load\n");
	strcpy(fullpath, basePath);
	strcat(fullpath, "audio/engine.wav");
	if (engine.load(fullpath) == -1)
		printf("Didn't load engine sound");


	GLMmodel *model;

	strcpy(fullpath, basePath);
	strcat(fullpath, "gfx/ship.obj");
	model = glmReadOBJ(fullpath);
	shipList = glmList(model, GLM_TEXTURE | GLM_SMOOTH);
	glmDelete(model);

	printf("Lists built\n");

	LoadTextures();

	if (nlvl)
		current.initialize(argv[nlvl]);
	else
		current.initialize("levels/level1.lvl");

	printf("Level loaded\n");


	SDL_ShowCursor(false);
	//go into the drawing loop
	SDL_Event evt;
	SDL_TimerID frameg_timer = SDL_AddTimer(1000, AddFrames, NULL);
	SDL_EnableKeyRepeat(0, 0);
	SDL_WM_GrabInput(SDL_GRAB_ON);
	g_oxygen = current.maxOxygen;
	g_fuel = current.maxFuel;
	printf("Entering message loop\n");
	trackTime = 0;
	g_time = SDL_GetTicks();
	// {{{ Ugly giant switch statements
	while (1) {
		if (SDL_PollEvent(&evt)) {
			if (evt.type == SDL_KEYDOWN) {
				sym = evt.key.keysym.sym;
				if (sym == SDLK_LEFT)
					keys[KEY_LEFT] = true;
				else if (sym == SDLK_RIGHT)
					keys[KEY_RIGHT] = true;
				else if (sym == SDLK_UP)
					keys[KEY_UP] = true;
				else if (sym == SDLK_DOWN)
					keys[KEY_DOWN] = true;
				else if (sym == SDLK_SPACE)
					keys[KEY_JUMP] = true;
				else if (sym == SDLK_e)
					/*keys[KEY_RAISE] = true; */
					state = EDITOR;
				else if (sym == SDLK_r)
					keys[KEY_LOWER] = true;
				else if (sym == SDLK_s)
					current.Save();
				else if (sym == SDLK_ESCAPE) {
					if (state == EDITOR)
						current.Sort();
					state = MENU;
				}
			} else if (evt.type == SDL_KEYUP) {
				sym = evt.key.keysym.sym;
				if (sym == SDLK_LEFT)
					keys[KEY_LEFT] = false;
				else if (sym == SDLK_RIGHT)
					keys[KEY_RIGHT] = false;
				else if (sym == SDLK_UP)
					keys[KEY_UP] = false;
				else if (sym == SDLK_DOWN)
					keys[KEY_DOWN] = false;
				else if (sym == SDLK_SPACE)
					keys[KEY_JUMP] = false;
				else if (sym == SDLK_e)
					keys[KEY_RAISE] = false;
				else if (sym == SDLK_r)
					keys[KEY_LOWER] = false;
			} else if (evt.type == SDL_MOUSEBUTTONDOWN) {
				if (evt.button.button == SDL_BUTTON_LEFT)
					g_LMouse = true;
				else if (evt.button.button ==
					 SDL_BUTTON_RIGHT)
					g_RMouse = true;
				else if (evt.button.button ==
					 SDL_BUTTON_MIDDLE)
					g_MMouse = true;
				else if (evt.button.button == 4) {	//mousewheel up
					printf("%i\n", e_curMaterial);
					e_curMaterial++;
					if (e_curMaterial == 10)
						e_curMaterial = 1;
				}
			} else if (evt.type == SDL_MOUSEBUTTONUP) {
				if (evt.button.button == SDL_BUTTON_LEFT)
					g_LMouse = false;
				else if (evt.button.button ==
					 SDL_BUTTON_RIGHT)
					g_RMouse = false;
				else if (evt.button.button ==
					 SDL_BUTTON_MIDDLE)
					g_MMouse = false;
			} else if (evt.type == SDL_MOUSEMOTION) {
				MouseMoved(evt.motion);
			} else if (evt.type == SDL_QUIT)
				break;
		}
		switch (state) {
		case MENU:
			mDrawScene();
			break;
		case GAME:
			gDrawScene();
			break;
		case EDITOR:
			gRunEditor();
			break;
		default:
			printf("Weird state; dying\n");
			return 0;
		}

		frames++;
		SDL_Delay(1);	//surrender the rest of the g_timeslice
	}
	// }}}
	SDL_WM_GrabInput(SDL_GRAB_OFF);
	SDL_ShowCursor(true);
	printf("Quitting SDL\n");
	SDL_RemoveTimer(frameg_timer);
	printf("Quitting g_timer subsystem\n");
	SDL_QuitSubSystem(SDL_INIT_TIMER);
	printf("Leaving Video subsystem\n");
	SDL_QuitSubSystem(SDL_INIT_VIDEO);
	printf("Leaving SDL\n");
	SDL_Quit();
	printf("All done\n");
	return 0;
}

void InitGL()
{
	glShadeModel(GL_SMOOTH);
	glCullFace(GL_BACK);
	glFrontFace(GL_CCW);
	glEnable(GL_CULL_FACE);
	glClearColor(0.5f, 0.5f, 0.5f, 1);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(100.0f, 800 / 600, 0.1f, 1000);
	glEnable(GL_TEXTURE_2D);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glEnable(GL_DEPTH_TEST);	// Enables Depth Testing
	glDepthFunc(GL_LEQUAL);	// The Type Of Depth Test To Do

}

void HandleKeys(double SPF, float *collisions, float floor)
{

	float strafemul = player.spdz > 15 ? 15 : player.spdz;	//because if it's allowed to run too high, the ship moves stupidly fast
	if (strafemul == 0)
		strafemul = 1;

	if (keys[KEY_UP]) {
		player.spdz += (SPF * 30.0f);
		if (engine.state != 1)
			engine.play(1);
	}
	if (keys[KEY_DOWN])
		player.spdz -= (SPF * 30.0f);

	if (keys[KEY_RIGHT] && player.animframe == 0 && (player.onground || player.offGroundTime < 0.3f)) {	//only have control if on the ground
		player.spdx += (1.0f * strafemul);
		if (player.spdx > strafemul)
			player.spdx = strafemul;
	}
	if (keys[KEY_LEFT] && player.animframe == 0
	    && (player.onground || player.offGroundTime < 0.3f)) {
		player.spdx -= (1.0f * strafemul);
		if (player.spdx < -strafemul)
			player.spdx = -strafemul;
	}

	if (keys[KEY_JUMP] && player.spdy == 0) {
		player.spdy = 60;
		player.y += 0.001f;
		player.onground = false;
		player.offGroundTime = 0;
	}

	if (player.spdz < 0) {
		player.spdz = 0;
		engine.pause();
	}
	if (player.spdz > 30)
		player.spdz = 30;

}

void updatePlayer(double SPF, float floor, float *collisions)
{
	if (collisions[FORWARD] != -1 && collisions[FORWARD] < 0.2f) {
		if (player.spdz > 10.0f) {	//we crashed fast
			engine.stop();
			explosion.play(0);
			explode(player);
			player.exploding = 11;
			player.spdy = 0;
		}
		player.spdz = 0;
	}

	g_oxygen -= SPF;
	g_fuel -= (player.spdz * SPF) / 10;

	if (g_oxygen < 0 || g_fuel < 0) {
		explode(player);
		player.exploding = 11;
	}

	player.z += SPF * player.spdz;

	player.spdy -= SPF * 200.0f / (500.0f/current.gravity);
	if (collisions[UP] == -1 || SPF * player.spdy < collisions[UP])
		player.y += SPF * player.spdy;
	else{
		player.y += collisions[UP] - 0.1f;
		player.spdy = 0;
	}
	if (player.y < floor) {
		player.y = floor;
		player.spdy = 0;
	}
	if (player.y == -15) {
		explode(player);
		player.exploding = 11;
		engine.stop();
		explosion.play(false);
	}

	if (player.spdx > 0 && collisions[RIGHT] != -1.0f)
		if (player.spdx * SPF >= collisions[RIGHT]) {
			player.x += collisions[RIGHT] - 0.01f;
			player.spdx = 0;
		}

	if (player.spdx < 0 && collisions[LEFT] != -1.0f)
		if (-(player.spdx * SPF) >= collisions[LEFT]) {
			player.x -= collisions[LEFT] - 0.01f;
			player.spdx = 0;
		}

	player.x += player.spdx * SPF;

	if (player.spdy == 0 && player.animframe == 0)	//friction
		player.spdx *= 0.1f * SPF;


	if (fabs(player.spdx) < 0.1f)
		player.spdx = 0;

	if (player.y == floor)
		player.onground = true;
	else
		player.onground = false;

}

void checkTerrain(double SPF, float floor, int surface)
{
	if (player.y <= floor) {
		switch (surface) {
		case -1:	//gap
			return;
		case 1:	//rtile1
			break;
		case 2:	//rtile2
			break;
		case 3:	//rtile3
			break;
		case 4:	//supply
			g_fuel = current.maxFuel;
			g_oxygen = current.maxOxygen;
			break;
		case 5:	//burning
			player.exploding = 11;
			player.spdz = 0;
			player.spdy = 0;
			engine.stop();
			explosion.play(false);
			explode(player);
			break;
		case 6:	//speedup
			player.spdz *= 1.1;
			break;	//slowdown
		case 7:
			player.spdz *= 0.9;
			break;
		case 8:	//slippy
			player.animframe = 1;
			break;
		case 9:
			player.spdz = 0;
			player.spdy = 0;
			player.exploding = 12;
			explode(player);
			engine.pause();
			finished = 1;
			break;
		}
	}
	return;
}

void drawMeters()
{

	//finally... draw g_throttle, g_fuel, O2
	glLoadIdentity();
	glEnable(GL_BLEND);
	glColor4f(0.1f, 0.39f, 0.82f, 0.5f);	//a nice electric blue
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
	// {{{ Bunch of colourful quads
	glBegin(GL_QUADS);
	glVertex3f(player.spdz / 30.0f, 1.0f, -1.0f);
	glVertex3f(-player.spdz / 30.0f, 1.0f, -1.0f);
	glVertex3f(-player.spdz / 30.0f, 0.9f, -1.0f);
	glVertex3f(player.spdz / 30.0f, 0.9f, -1.0f);
	glEnd();		//g_fuel and O2 can be added in quite easily...

	//next: Oxygen - left
	//Oxygen - erm... green
	glColor4f(0.0f, 1.0f, 0.0f, 0.5f);
	glBegin(GL_QUADS);
	glVertex3f(-1.2f, -g_oxygen / current.maxOxygen, -1.0f);
	glVertex3f(-1.1f, -g_oxygen / current.maxOxygen, -1.0f);
	glVertex3f(-1.1f, g_oxygen / current.maxOxygen, -1.0f);
	glVertex3f(-1.2f, g_oxygen / current.maxOxygen, -1.0f);
	glEnd();
	//g_fuel, in yellow. I really should precalc the values...
	glColor4f(1.0f, 1.0f, 0.0f, 0.5f);
	glBegin(GL_QUADS);
	glVertex3f(1.1f, -g_fuel / current.maxFuel, -1.0f);
	glVertex3f(1.2f, -g_fuel / current.maxFuel, -1.0f);
	glVertex3f(1.2f, g_fuel / current.maxFuel, -1.0f);
	glVertex3f(1.1f, g_fuel / current.maxFuel, -1.0f);
	glEnd();

	if (!player.onground) {
		glColor4f(1.0f, 0.1f, 0.1f, 0.5f);
		glBegin(GL_QUADS);
		glVertex3f(1.1f, -1.1f, -1.0f);
		glVertex3f(1.1f, -1.2f, -1.0f);
		glVertex3f(1.2f, -1.2f, -1.0f);
		glVertex3f(1.2f, -1.1f, -1.0f);
		glEnd();
	}
	// }}}
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);

	return;
}

void checkCollisions(float collisions[], float *floor, int *surface)
{

	float tmpfront = player.z + 0.1f, tmpleft =
	    player.x - 1.9f, tmpright = player.x + 1.9f, tmpback =
	    player.z - 5.0f, tmptop = player.y + 1.0f, dist = 0;
	// {{{ Collision detection
	lvlpoint *curBlock = current.head;
	while (curBlock != NULL) {
		if (collisions[FORWARD] != -1
		    && curBlock->blz > player.z + collisions[FORWARD]) {
			curBlock = curBlock->next;
			continue;
		}
		if (curBlock->trz < tmpback) {
			curBlock = curBlock->next;
			continue;
		}
		//right
		if (curBlock->trz > tmpback
		    && curBlock->blz < tmpfront
		    && curBlock->blx > tmpright
		    && curBlock->trY > player.y
		    && curBlock->bly <= player.y) {
			dist = curBlock->blx - tmpright;
			if (collisions[RIGHT] == -1
			    || dist < collisions[RIGHT])
				collisions[RIGHT] = dist;
		}
		//to the left
		if (curBlock->trz > tmpback
		    && curBlock->blz < tmpfront
		    && curBlock->trx < tmpleft
		    && curBlock->trY > player.y
		    && curBlock->bly <= player.y) {
			dist = tmpleft - curBlock->trx;
			if (collisions[LEFT] == -1
			    || dist < collisions[LEFT])
				collisions[LEFT] = dist;
		}
		//front
		if (curBlock->blx < tmpright && curBlock->trx > tmpleft
		    && curBlock->blz > tmpfront
		    && curBlock->trY >= player.y + 0.01f
		    && curBlock->bly <= player.y) {
			dist = curBlock->blz - tmpfront;
			if (collisions[FORWARD] == -1
			    || dist < collisions[FORWARD])
				collisions[FORWARD] = dist;
		}

		if (curBlock->trx > tmpleft && curBlock->blx < tmpright && curBlock->trz > tmpback && curBlock->blz < tmpfront - 0.1f) {	//the 0.1 prevents the ship from 'lifting' itself
			if (curBlock->trY <= player.y){
				if (*floor < curBlock->trY)
					*floor = curBlock->trY;
				*surface = curBlock->material;
				if (curBlock->material == 8)	/*slippy */
					player.animframe = 1;
			} else {//above
				if (collisions[UP] == -1 || curBlock->bly - tmptop < collisions[UP])
					collisions[UP] = curBlock->bly - tmptop;
			}			
		}
		curBlock = curBlock->next;
	}
	// }}}
}

void drawBackground()
{

	glDisable(GL_DEPTH_TEST);
	glBegin(GL_QUADS);

	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(-1.0f, -1.0f);
	glTexCoord2f(1.0f, 0.0f);
	glVertex2f(1.0f, -1.0f);
	glTexCoord2f(1.0f, 1.0f);
	glVertex2f(1.0f, 1.0f);
	glTexCoord2f(0.0f, 1.0f);
	glVertex2f(-1.0f, 1.0f);

	glEnd();
	glEnable(GL_DEPTH_TEST);
	return;
}

void gDrawScene()
{
	float floor = -15;
	int surface = -1;
	Uint32 g_timeDiff = 0;
	double SecsperFrame = 0;
	lastCycle = currentTime;
	currentTime = SDL_GetTicks();
	GLenum retval;
	//this needs to be first
	while (!SecsperFrame) {
		g_timeDiff = SDL_GetTicks() - g_time;
		SecsperFrame = (double) g_timeDiff / 1000.0;
	}

//      SecsperFrame = (currentTime - lastCycle) / 1000.0f;
	SecsperFrame = (SDL_GetTicks() - lastCycle) / 1000.0f;
	player.onground ? SecsperFrame *= 0.9f : SecsperFrame *= 0.86f;	//seems to give a more skyroads feel...
	g_throttle += (currentTime - lastCycle);
	if (!finished)
		trackTime += (currentTime - lastCycle);

	player.offGroundTime += SecsperFrame;

	//if one ms hasn't passed, wait until one has (effectively limiting to 1000 fps)
	glMatrixMode(GL_MODELVIEW);
	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
	glLoadIdentity();

	glFrontFace(GL_CCW);
	glEnable(GL_TEXTURE_2D);
	glColor3f(1.0f, 1.0f, 1.0f);
	glBindTexture(GL_TEXTURE_2D, textures[13]);

	glTranslatef(0.0f, 0.0f, -0.5f);

	drawBackground();

	retval = glGetError();
	if (retval != GL_NO_ERROR) {
		printf("OpenGL Error received!!! (setup)... %i \n",
		       retval);
	}
	glLoadIdentity();


	retval = glGetError();
	if (retval != GL_NO_ERROR) {
		printf("OpenGL Error received!!! (background)... %i \n",
		       retval);
	}

	glRotatef(6.0f, 1.0f, 0.0f, 0.0f);
	glTranslatef(0.0f, -14.0f, player.z - 19);	//drop the camera back just enough to not lose the ship out the side...

	float collisions[4] = { -1.0f, -1.0f, -1.0f, -1.0f };
	player.animframe = 0;
	checkCollisions(collisions, &floor, &surface);

	if (player.exploding == false && trackTime > 0) {
		updatePlayer(SecsperFrame, floor, collisions);
		checkTerrain(SecsperFrame, floor, surface);
		HandleKeys(SecsperFrame, collisions, floor);
	} else if (player.exploding) {
		if (!updateParticles(SecsperFrame)) {
			if (player.exploding == 12)	//load the next level
				if (!current.loadNext())
					state = MENU;
			printf("Your g_time was %f seconds\n",
			       trackTime / 1000.0f);
			Reset();
			trackTime = -1500; //the countdown timer
			return;
		}
	}

	//draw the level
	glColor3f(0.0f, 0.0f, 0.0f);
	//render the level
	glColor3f(1.0f, 1.0f, 1.0f);
	current.Render(player.z + 70.0f, player.z - 15.0f);

	retval = glGetError();
	if (retval != GL_NO_ERROR) {
		printf("OpenGL Error received!!! (level)... %i \n",
		       retval);
	}
	//draw the 'player'
	if (!player.exploding) {
		glColor3f(1.0f, 1.0f, 1.0f);
		glPushMatrix();
		glTranslatef(player.x, player.y, -player.z);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, textures[0]);
		glCallList(shipList);
		glDisable(GL_TEXTURE_2D);
		glFrontFace(GL_CCW);
		glPopMatrix();
	} else {
		glPushMatrix();
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glEnable(GL_BLEND);
		glBindTexture(GL_TEXTURE_2D, textures[player.exploding]);
		glTranslatef(player.x, player.y, -player.z);
		drawParticles(SecsperFrame);
		glDisable(GL_BLEND);
		glPopMatrix();
	}

	retval = glGetError();
	if (retval != GL_NO_ERROR) {
		printf("OpenGL Error received!!! (player)... %i \n",
		       retval);
	}

	drawMeters();

	glEnable(GL_TEXTURE_2D);

	glBindTexture(GL_TEXTURE_2D, textures[14]);
	glLoadIdentity();
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);	//the blend is just to make the black disappear
	glTranslatef(-2.0f, 10.0f, -10.0f);

	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);

//      drawTime();
	text.drawf("%2.2f", trackTime / 1000.0f);

	glColor4f(0.5f, 1.0f, 0.5f, 1.0f);
	glLoadIdentity();
	glTranslatef(7.0f, 10.0f, -10.0f);
	text.drawf("%i", fps);
	glLoadIdentity();
	glColor4f(0.5f, 0.75f, 1.0f, 1.0f);
	glTranslatef(-10.0f, -10.0f, -10.0f);
	text.drawf("%i", current.gravity);

	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);

	if (trackTime < 0){
		glLoadIdentity();
		glDisable(GL_TEXTURE_2D);
		glEnable(GL_BLEND);
		glColor4f(1.0f, 1.0f, 1.0f, (-trackTime)/1500.0f);
		glBegin(GL_QUADS);
		glVertex3f(-1.5f, -1.5f, -1.0f);
		glVertex3f( 1.5f, -1.5f, -1.0f);
		glVertex3f( 1.5f,  1.5f, -1.0f);
		glVertex3f(-1.5f,  1.5f, -1.0f);
		glEnd();
		glDisable(GL_BLEND);
		glEnable(GL_TEXTURE_2D);
	}

	SDL_GL_SwapBuffers();

	retval = glGetError();
	if (retval != GL_NO_ERROR) {
		printf("OpenGL Error received!!! (counters)... %i \n",
		       retval);
	}


	
	g_time = SDL_GetTicks();
}

void MouseMoved(SDL_MouseMotionEvent motion)
{
	cursor.x = (motion.x - 400) * 0.03;

	if (!g_MMouse)
		if (stVert)
			cursor.y = (600 - motion.y) * 0.10f - 3.0f;
		else
			cursor.z = (600 - motion.y) * 0.10f;
	else
		g_camPos += motion.yrel;

	return;
}

void gRunEditor()
{
	glMatrixMode(GL_MODELVIEW);
	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
	glLoadIdentity();

	glFrontFace(GL_CCW);
	glEnable(GL_TEXTURE_2D);
	glColor3f(1.0f, 1.0f, 1.0f);

	glRotatef(10.0f, 1.0f, 0.0f, 0.0f);
	glTranslatef(0.0f, -16.0f, -15.0f);



	glTranslatef(0.0f, 0.0f, g_camPos);
	glEnable(GL_TEXTURE_2D);
	current.Render(-100.0f, 1000.0f);
	//If we're to highlight a block, do it now
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glColor4f(1.0f, 1.0f, 1.0f, 0.5f);
	lvlpoint *tmp = current.head;
	e_edSelected = NULL;
	while (tmp != NULL) {
		if (cursor.x * 2 > tmp->blx && cursor.x * 2 < tmp->trx &&
		    cursor.y > tmp->bly && cursor.y < tmp->trY &&
		    cursor.z + g_camPos > tmp->blz
		    && cursor.z + g_camPos < tmp->trz) {
			e_edSelected = tmp;
			// {{{ Another big ugly box

			glPushMatrix();
			glDisable(GL_TEXTURE_2D);
			glScalef(1.1f, 1.1f, 1.0f);
			glTranslatef(0.0f, 0.01f, 0.01f);
			glBegin(GL_QUADS);
			//first quad: back face (visible from rear)
			glVertex3f(tmp->blx, tmp->bly, -tmp->blz);
			glVertex3f(tmp->trx, tmp->bly, -tmp->blz);
			glVertex3f(tmp->trx, tmp->trY, -tmp->blz);
			glVertex3f(tmp->blx, tmp->trY, -tmp->blz);

			//right face
			glVertex3f(tmp->trx, tmp->bly, -tmp->blz);
			glVertex3f(tmp->trx, tmp->bly, -tmp->trz);
			glVertex3f(tmp->trx, tmp->trY, -tmp->trz);
			glVertex3f(tmp->trx, tmp->trY, -tmp->blz);

			//left face
			glVertex3f(tmp->blx, tmp->bly, -tmp->blz);
			glVertex3f(tmp->blx, tmp->trY, -tmp->blz);
			glVertex3f(tmp->blx, tmp->trY, -tmp->trz);
			glVertex3f(tmp->blx, tmp->bly, -tmp->trz);

			//top face
			glVertex3f(tmp->blx, tmp->trY, -tmp->blz);
			glVertex3f(tmp->trx, tmp->trY, -tmp->blz);
			glVertex3f(tmp->trx, tmp->trY, -tmp->trz);
			glVertex3f(tmp->blx, tmp->trY, -tmp->trz);

			glEnd();
			glPopMatrix();
			// }}}
		}
		tmp = tmp->next;

	}
	glDisable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, textures[0]);
	glCallList(shipList);


	if (g_LMouse && editorState == 0) {
		//fill in blx, although we may have to swap around some components later
		stVert = 0;
		curPoint.blx = cursor.x;
		curPoint.blz = cursor.z + g_camPos;
		curPoint.bly = cursor.y;
		g_LMouse = false;
		editorState++;
	} else if (g_LMouse && editorState == 1) {
		//second part...
		curPoint.trz = cursor.z + g_camPos;
		g_LMouse = false;
		stVert = 1;
		editorState++;
	} else if (g_LMouse && editorState == 2) {
		//er... we should just add it
		g_LMouse = false;
		float spare;
		curPoint.trx = cursor.x;	//do it again, for no good reason
		curPoint.trY = cursor.y;

		curPoint.blx = roundf(curPoint.blx);
		curPoint.bly = roundf(curPoint.bly);
		curPoint.blz = roundf(curPoint.blz / 3) * 3;

		curPoint.trx = roundf(curPoint.trx);
		curPoint.trY = roundf(curPoint.trY);
		curPoint.trz = roundf(curPoint.trz / 3) * 3;

		curPoint.blx *= 3;
		curPoint.bly *= 3;

		curPoint.trx *= 3;
		curPoint.trY *= 3;

		//organise the points
		if (curPoint.blx > curPoint.trx) {
			spare = curPoint.blx;
			curPoint.blx = curPoint.trx;
			curPoint.trx = spare;
		}
		if (curPoint.bly > curPoint.trY) {
			spare = curPoint.bly;
			curPoint.bly = curPoint.trY;
			curPoint.trY = spare;
		}
		if (curPoint.blz > curPoint.trz) {
			spare = curPoint.blz;
			curPoint.trz = curPoint.blz;
			curPoint.blz = spare;
		}

		curPoint.material = e_curMaterial;
		printf("%f %f %f : %f %f %f\n", curPoint.blx, curPoint.bly,
		       curPoint.blz, curPoint.trx, curPoint.trY,
		       curPoint.trz);
		current.Add(&curPoint);
		editorState = 0;
		memset(&curPoint, 0, sizeof(lvlpoint));
		stVert = 0;
	}
	//add in shit for a fourth state - moving a corner

	if (g_RMouse == 1 && editorState == 0 && e_edSelected == 0) {
		stVert = !stVert;
		g_RMouse = false;
	} else if (g_RMouse == 1 && editorState == 0 && e_edSelected != NULL) {
		g_RMouse = false;
		current.Delete(e_edSelected);
	}

/*	if (g_MMouse == 1){
		g_MMouse = 0;
		cursor.y = 0;
	}
*/
	if (g_RMouse == 1 && g_LMouse == 1) {
		g_RMouse = g_MMouse = 0;
		cursor.y = 0;
	}

	glPushMatrix();
	glDisable(GL_TEXTURE_2D);
	glColor3f(1.0f, 1.0f, 1.0f);

	if (editorState == 2)
		glTranslatef(cursor.x * 3, cursor.y * 3.0f, -curPoint.trz);
	else
		glTranslatef(cursor.x * 3, cursor.y * 3.0f,
			     -cursor.z - g_camPos);

	// {{{ The cursor
	glBegin(GL_QUADS);
	glVertex3f(-0.2f, 0, 0);
	glVertex3f(0.2f, 0, 0);
	glVertex3f(0.2f, 1, 0);
	glVertex3f(-0.2f, 1, 0);
	glEnd();

	glPopMatrix();

	glBegin(GL_LINES);
	glVertex3f(floorf(cursor.x) * 3, 0, -cursor.z - g_camPos - 8);
	glVertex3f(floorf(cursor.x) * 3, 0, -cursor.z - g_camPos + 8);

	glVertex3f(ceilf(cursor.x) * 3, 0, -cursor.z - g_camPos - 8);
	glVertex3f(ceilf(cursor.x) * 3, 0, -cursor.z - g_camPos + 8);

	glVertex3f(floorf(cursor.x - 1.0f) * 3, 0,
		   -cursor.z - g_camPos - 8);
	glVertex3f(floorf(cursor.x - 1.0f) * 3, 0,
		   -cursor.z - g_camPos + 8);
	glVertex3f(ceilf(cursor.x + 1.0f) * 3, 0,
		   -cursor.z - g_camPos - 8);
	glVertex3f(ceilf(cursor.x + 1.0f) * 3, 0,
		   -cursor.z - g_camPos + 8);

	glVertex3f((cursor.x - 3) * 3, 0,
		   -floorf((cursor.z + g_camPos) / 3) * 3);
	glVertex3f((cursor.x + 3) * 3, 0,
		   -floorf((cursor.z + g_camPos) / 3) * 3);

	glVertex3f((cursor.x - 3) * 3, 0,
		   -ceilf((cursor.z + g_camPos) / 3) * 3);
	glVertex3f((cursor.x + 3) * 3, 0, -ceilf((cursor.z + g_camPos) / 3) * 3);	// /3*3 changes the resolution of floor()

	glVertex3f((cursor.x - 3) * 3, 0,
		   -floorf((cursor.z + g_camPos) / 3 - 1.0f) * 3);
	glVertex3f((cursor.x + 3) * 3, 0,
		   -floorf((cursor.z + g_camPos) / 3 - 1.0f) * 3);

	glVertex3f((cursor.x - 3) * 3, 0,
		   -ceilf((cursor.z + g_camPos) / 3 + 1.0f) * 3);
	glVertex3f((cursor.x + 3) * 3, 0,
		   -ceilf((cursor.z + g_camPos) / 3 + 1.0f) * 3);

	glEnd();


	// }}}


	//draw current tile
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, textures[e_curMaterial]);
	glPushMatrix();
	glLoadIdentity();
	glTranslatef(0.0f, 0.0f, -1.0f);

	// {{{ A box
	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(-1.0f, 0.8f);
	glTexCoord2f(1.0f, 0.0f);
	glVertex2f(-0.9f, 0.8f);
	glTexCoord2f(1.0f, 1.0f);
	glVertex2f(-0.9f, 1.0f);
	glTexCoord2f(0.0f, 1.0f);
	glVertex2f(-1.0f, 1.0f);
	glEnd();
	// }}}
	glDisable(GL_TEXTURE_2D);
	glPopMatrix();

	lvlpoint tmppoint;	//I know the second p should be capitalized, but it's such a pain to type
	switch (editorState) {
	case 1:
		tmppoint.blx = curPoint.blx * 3;
		tmppoint.bly = curPoint.bly * 3;
		tmppoint.blz = curPoint.blz;

		tmppoint.trx = cursor.x * 3;
		tmppoint.trY = cursor.y * 3.0f;
		tmppoint.trz = cursor.z + g_camPos;
		break;
	case 2:
		tmppoint.blx = curPoint.blx * 3;
		tmppoint.bly = curPoint.bly * 3;
		tmppoint.blz = curPoint.blz;

		tmppoint.trx = cursor.x * 3;
		tmppoint.trY = cursor.y * 3.0f;
		tmppoint.trz = curPoint.trz;
	}
	if (editorState != 0) {
		glEnable(GL_BLEND);
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_CULL_FACE);
		glColor4f(1.0f, 1.0f, 1.0f, 0.3f);
		/// {{{ Boxes and stuff         
		glBegin(GL_QUADS);
		//front face
		glVertex3f(tmppoint.blx, tmppoint.bly, -tmppoint.blz);
		glVertex3f(tmppoint.trx, tmppoint.bly, -tmppoint.blz);
		glVertex3f(tmppoint.trx, tmppoint.trY, -tmppoint.blz);
		glVertex3f(tmppoint.blx, tmppoint.trY, -tmppoint.blz);
		//top face
		glVertex3f(tmppoint.blx, tmppoint.trY, -tmppoint.blz);
		glVertex3f(tmppoint.trx, tmppoint.trY, -tmppoint.blz);
		glVertex3f(tmppoint.trx, tmppoint.trY, -tmppoint.trz);
		glVertex3f(tmppoint.blx, tmppoint.trY, -tmppoint.trz);

		//bottom face
		glVertex3f(tmppoint.blx, tmppoint.bly, -tmppoint.blz);
		glVertex3f(tmppoint.trx, tmppoint.bly, -tmppoint.blz);
		glVertex3f(tmppoint.trx, tmppoint.bly, -tmppoint.trz);
		glVertex3f(tmppoint.blx, tmppoint.bly, -tmppoint.trz);

		//left face
		glVertex3f(tmppoint.blx, tmppoint.bly, -tmppoint.blz);
		glVertex3f(tmppoint.blx, tmppoint.trY, -tmppoint.blz);
		glVertex3f(tmppoint.blx, tmppoint.trY, -tmppoint.trz);
		glVertex3f(tmppoint.blx, tmppoint.bly, -tmppoint.trz);

		//right face
		glVertex3f(tmppoint.trx, tmppoint.bly, -tmppoint.blz);
		glVertex3f(tmppoint.trx, tmppoint.bly, -tmppoint.trz);
		glVertex3f(tmppoint.trx, tmppoint.trY, -tmppoint.trz);
		glVertex3f(tmppoint.trx, tmppoint.trY, -tmppoint.blz);

		glEnd();
		//now, put in the wireframe lines, to make the box properly visible
		glDisable(GL_BLEND);

		glColor3f(0.0f, 0.0f, 0.0f);

		glBegin(GL_LINES);

		//top face
		glVertex3f(tmppoint.blx, tmppoint.trY, -tmppoint.trz);
		glVertex3f(tmppoint.trx, tmppoint.trY, -tmppoint.trz);

		glVertex3f(tmppoint.blx, tmppoint.trY, -tmppoint.trz);
		glVertex3f(tmppoint.blx, tmppoint.trY, -tmppoint.blz);

		glVertex3f(tmppoint.trx, tmppoint.trY, -tmppoint.trz);
		glVertex3f(tmppoint.trx, tmppoint.trY, -tmppoint.blz);

		//frame the front face

		glVertex3f(tmppoint.blx, tmppoint.trY, -tmppoint.blz);
		glVertex3f(tmppoint.trx, tmppoint.trY, -tmppoint.blz);

		glVertex3f(tmppoint.blx, tmppoint.trY, -tmppoint.blz);
		glVertex3f(tmppoint.blx, tmppoint.bly, -tmppoint.blz);

		glVertex3f(tmppoint.trx, tmppoint.trY, -tmppoint.blz);
		glVertex3f(tmppoint.trx, tmppoint.bly, -tmppoint.blz);

		glVertex3f(tmppoint.blx, tmppoint.bly, -tmppoint.blz);
		glVertex3f(tmppoint.trx, tmppoint.bly, -tmppoint.blz);

		//bottom edges
		glVertex3f(tmppoint.blx, tmppoint.bly, -tmppoint.blz);
		glVertex3f(tmppoint.blx, tmppoint.bly, -tmppoint.trz);

		glVertex3f(tmppoint.trx, tmppoint.bly, -tmppoint.blz);
		glVertex3f(tmppoint.trx, tmppoint.bly, -tmppoint.trz);

		//remaining 3, at the back

		glVertex3f(tmppoint.blx, tmppoint.bly, -tmppoint.trz);
		glVertex3f(tmppoint.trx, tmppoint.bly, -tmppoint.trz);

		glVertex3f(tmppoint.blx, tmppoint.bly, -tmppoint.trz);
		glVertex3f(tmppoint.blx, tmppoint.trY, -tmppoint.trz);

		glVertex3f(tmppoint.trx, tmppoint.bly, -tmppoint.trz);
		glVertex3f(tmppoint.trx, tmppoint.trY, -tmppoint.trz);

		glEnd();
		// }}}
		glEnable(GL_CULL_FACE);
		glEnable(GL_DEPTH_TEST);
		glDisable(GL_BLEND);

	}

	glLoadIdentity();
	glTranslatef(-10.0f, -10.0f, -10.0f);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, textures[14]);
	text.drawf("%.1f %.1f %.1f", cursor.x, cursor.y, cursor.z);
	glDisable(GL_BLEND);


	SDL_GL_SwapBuffers();
	return;
}


bool LoadTexture(char *filename, int i)
{
	SDL_Surface *bgrText;
	SDL_Rect area;
	Uint32 saved_flags;
	Uint8 saved_alpha;
	SDL_Surface *image;
	char fullpath[64];

	strcpy(fullpath, basePath);
	strcat(fullpath, filename);


	image = IMG_Load(fullpath);

	if (image == NULL) {
		printf("Here comes a segfault... IMG_Load %i failed: %s\n",
		       i, SDL_GetError());
		return false;
	}

	printf("Loading %s into place %i\n", fullpath, i);

	bgrText =
	    SDL_CreateRGBSurface(SDL_HWSURFACE, image->w, image->h,
				 image->format->BitsPerPixel, 0x000000FF,
				 0x0000FF00, 0x00FF0000, 0xFF000000);
	area.x = 0;
	area.y = 0;
	area.w = image->w;
	area.h = image->h;
	saved_flags = image->flags & (SDL_SRCALPHA | SDL_RLEACCELOK);
	saved_alpha = image->format->alpha;
	if ((saved_flags & SDL_SRCALPHA) == SDL_SRCALPHA)
		SDL_SetAlpha(image, 0, 0);
	SDL_BlitSurface(image, &area, bgrText, &area);
	if ((saved_flags & SDL_SRCALPHA) == SDL_SRCALPHA)
		SDL_SetAlpha(image, saved_flags, saved_alpha);

	glBindTexture(GL_TEXTURE_2D, textures[i]);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, image->w, image->h, 0,
		     GL_RGB, GL_UNSIGNED_BYTE, bgrText->pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);	// Linear Filtering
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	GLenum retval = glGetError();
	if (retval != GL_NO_ERROR) {
		printf
		    ("OpenGL Error received!!! (loading texture %i)... %i \n",
		     i, retval);
	}

	SDL_FreeSurface(image);
	SDL_FreeSurface(bgrText);
	return true;
}

void LoadTextures()
{

	char files[TEXTURES][32] = { {"gfx/skin.png"}, {"gfx/rtile1.png"},
	{"gfx/rtile2.png"},
	{"gfx/rtile3.png"}, {"gfx/supply.png"}, {"gfx/burning.png"},
	{"gfx/speedup.png"}, {"gfx/sticky.png"}, {"gfx/slippery.png"},
	{"gfx/chequered.png"}, {"gfx/menubg.png"}, {"gfx/boom.png"},
	{"gfx/goal.png"}, {"gfx/menubg.png"}, {"gfx/font.png"}
	};
	printf("Loading textures...\n");
	glEnable(GL_TEXTURE_2D);
	glGenTextures(TEXTURES, &textures[0]);
	int i;
	for (i = 0; i < TEXTURES; i++)
		LoadTexture(files[i], i);
	LoadTexture("gfx/skin.png", 0);	//this seems necessary for it to work under windoze... weird...
}
